{#
Renders a ``div`` to display a Bokeh model into a Jupyter Notebook.

:param plot_script: typically the output of PLOT_SCRIPT
:type plot_script: str

:param plot_div: typically the output of PLOT_DIV
:type plot_div: str

#}
{{ plot_div|indent(4) }}
<script type="text/javascript">
  {{ plot_script|indent(2) }}
</script>
